#!/usr/bin/env python3

"""Draw graphs from the database."""

import click
import itertools
import json
import pathlib
import re
import typing

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np


class DynAnnot:
    """Create dynamic annotations."""

    def __call__(self, event):
        # print(event)
        # vis = annot.get_visible()
        pass


def get_floats(value: float | int | str | list | dict) -> typing.Iterable[float]:
    """Extract all the number recursively."""
    match value:
        case float():
            yield value
        case int():
            yield float(value)
        case str():
            try:
                yield float(value)
            except ValueError:
                pass
        case list() | tuple() | set() | frozenset():
            yield from (e for v in value for e in get_floats(v))
        case dict():
            yield from (e for v in value.values() for e in get_floats(v))
        case _:
            raise TypeError(f"unrecognize value {value}")


def get_float_mean(value: object) -> float:
    """Return the mean of the float, or raise ValueError."""
    if len(values := list(get_floats(value))) == 0:
        raise ValueError(f"no number found in {value}")
    return float(np.mean(values))


def get_value(value: dict | list, field: str | None) -> object:
    """Exctract the value that correspond to the given field.

    Returns None if it failed.
    """
    if field is None:
        return None
    assert isinstance(value, dict | list), value.__class__.__name__
    assert isinstance(field, str), field.__class__.__name__
    if isinstance(value, list):
        return [get_value(e, field) for e in value]
    key, *sub_keys = field.split(".")
    if (value := value.get(key, None)) is not None:
        if sub_keys:
            return get_value(value, ".".join(sub_keys))
    return value


def _log_is_better(data_lin: np.ndarray[np.float64]) -> bool:
    """Return True if the data should be plot in a log scale."""
    if data_lin.min() <= 0.0:
        return False
    if data_lin.max() / data_lin.min() < 10.0**1.5:  # minimum 1.5 decades
        return False
    data_log = np.log(data_lin)
    bins = 1 + round(len(data_lin)**0.5)
    hist_lin, _ = np.histogram(data_lin, bins=bins)
    hist_lin = (hist_lin.astype(np.float64) + np.finfo(np.float64).eps) / len(data_lin)
    hist_log, _ = np.histogram(data_log, bins=bins)
    hist_log = (hist_log.astype(np.float64) + np.finfo(np.float64).eps) / len(data_log)
    kullback_lin = -np.log(hist_lin).sum()  # rigouresly np.sum(np.log((1/bins) / hist_lin)) / bins
    kullback_log = -np.log(hist_log).sum()
    return kullback_log < kullback_lin


def _draw(axes: plt.Axes, data: dict[str: dict], values: dict[str: dict[str]]):
    """Create the plot."""
    # set title
    if (lbl := set(map(str, values["w"].values()))) != {"None"}:
        axes.set_title(", ".join(sorted(lbl)))

    # set legend
    colors_label = sorted(set(map(str, values["c"].values())))
    shapes_label = sorted(set(map(str, values["s"].values())))
    if len(colors_label) > 1:
        for i in range(len(colors_label)):
            axes.scatter([], [], color=cm.rainbow(i/(len(colors_label)-1)), label=colors_label[i], marker=1)
    if len(shapes_label) > 1:
        for j in range(len(shapes_label)):
            axes.scatter([], [], color="black", label=shapes_label[j], marker=".x^1s+Dv234<>hH"[j%15])
    if len(colors_label) > 1 or len(shapes_label) > 1:
        axes.legend()

    # draw scatter cloud
    for i, j in itertools.product(range(len(colors_label)), range(len(shapes_label))):
        x_data, y_data = zip(*(
            (values["x"][name], values["y"][name]) for name in data
            if str(values["c"][name]) == colors_label[i] and str(values["s"][name]) == shapes_label[j]
        ))
        x_data, y_data = np.asarray(x_data, dtype=np.float64), np.asarray(y_data, dtype=np.float64)
        axes.scatter(
            x_data,
            y_data,
            color=cm.rainbow(i/(len(colors_label)-1)) if len(colors_label) > 1 else None,
            marker=".x^1s+Dv234<>hH"[j%15] if len(shapes_label) > 1 else None,
        )

    # put decoration
    axes.grid()
    x_data = np.asarray(list(values["x"].values()), dtype=np.float64)
    axes.set_xscale("log" if _log_is_better(x_data) else "linear")
    y_data = np.asarray(list(values["y"].values()), dtype=np.float64)
    axes.set_yscale("log" if _log_is_better(y_data) else "linear")
    # axes.set_xlabel(kwargs["x"])
    # axes.set_ylabel(kwargs["y"])


@click.command()
@click.argument("filename", required=True, type=click.Path())
@click.option("-x", required=True, help="the field along x axis")
@click.option("-y", required=True, help="the field along y axis")
@click.option("-c", "--color", type=str, help="criteria for points color")
@click.option("-s", "--shape", type=str, help="criteria for the points shape")
@click.option("-w", "--window", type=str, help="criteria for the points window")
@click.option("-u", "--unfold", type=click.Choice(list("xycsw")), multiple=True, help="the field to unfold")
@click.option("-f", "--filter", type=str, help="criteria to be filtered")
def main(filename: str, **kwargs):
    """Draw graphs from the database.

    Parameters
    ----------
    filename : pathlike
        The name of the json file.
    x, y : str
        The name of the field to plot for the x and the y axis.
        For example ``encode_power.energy``.
    filter : str, optional
        The condition to keep the file, formatted as ``field=value``.
        For example ``encoder=libx264``
    """
    # extract the database content
    filename = pathlib.Path(filename).expanduser()
    assert filename.is_file(), f"the given file {filename} doesn't exists"
    with open(filename, "rb") as raw:
        data = json.load(raw)

    # filter
    if kwargs.get("filter", ""):
        key, *pattern = kwargs["filter"].split("=")
        pattern = "=".join(pattern)
        plt.title(f"{key} = {pattern}")
        if not (data := {
            name: values for name, values in data.items()
            if re.search(pattern, get_value(values, key) or "") is not None
        }):
            raise KeyError(f"the filter {key} = {pattern} reject everything")

    # select values
    values = {
        "x": {n: get_value(v, kwargs["x"]) for n, v in data.items()},
        "y": {n: get_value(v, kwargs["y"]) for n, v in data.items()},
        "c": {n: get_value(v, kwargs["color"]) for n, v in data.items()},
        "s": {n: get_value(v, kwargs["shape"]) for n, v in data.items()},
        "w": {n: get_value(v, kwargs["window"]) for n, v in data.items()},
    }

    # unfold values
    for field in kwargs["unfold"]:
        for name in data.copy():
            if not isinstance(values[field][name], list | tuple):
                continue
            for i in range(len(values[field][name])):
                new_name = f"x{i+1}_{name}"
                for field_ in "xycsw":
                    if isinstance(values[field_][name], list | tuple):
                        values[field_][new_name] = values[field_][name][i]
                    else:
                        values[field_][new_name] = values[field_][name]
                data[new_name] = data[name]
            del data[name]
            for field_ in "xycsw":
                del values[field_][name]

    # convert to scalar
    values["x"] = {n: get_float_mean(v) for n, v in values["x"].items()}
    values["y"] = {n: get_float_mean(v) for n, v in values["y"].items()}

    # display
    fig = plt.figure(figsize=(16, 9), dpi=None)
    fig.suptitle(filename.stem)
    dyn_annot = DynAnnot()
    windows_label = sorted(set(map(str, values["w"].values())))
    for i, lbl in enumerate(windows_label):
        axes = fig.add_subplot(1, len(windows_label), i+1)
        axes.set_xlabel(kwargs["x"])
        axes.set_ylabel(kwargs["y"])
        selected = {n for n, e in values["w"].items() if str(e) == lbl}
        _draw(
            axes,
            {n: d for n, d in data.items() if n in selected},
            {f: {n: e for n, e in v.items() if n in selected} for f, v in values.items()},
        )
    plt.tight_layout()
    lbl = " ".join(f"{k}={kwargs[k]}" for k in sorted(kwargs) if kwargs[k])
    plt.savefig(
        (pathlib.Path.cwd() / f"{filename.stem} {lbl}").with_suffix(".svg"),
        # transparent=True,
        format="svg",
    )
    fig.canvas.mpl_connect("motion_notify_event", dyn_annot)
    plt.show()


    # # display
    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.set_title("Interactive MENDEVI plot")
    # scenes = []

    # if len(names := x_values.keys() & y_values.keys()) == 0:
    #     names = {k for v in data.values() for k in v}
    #     raise KeyError(f"no x={kwargs['x']} and y={kwargs['y']} label found, only {sorted(names)}")
    # names, x_data, y_data = zip(
    #     *((k, x_values[k], y_values[k]) for k in names)
    # )
    # x_data, y_data = np.asarray(x_data, np.float64), np.asarray(y_data, np.float64)
    # if "color" in kwargs:
    #     colors = autocolors()
    #     colors = {}
    #     for i, name in enumerate(names):
    #         cat = data[name].get(kwargs["color"], "not specified")
    #         colors[cat] = colors.get(cat, [])
    #         colors[cat].append(i)
    #     for cat in sorted(colors):
    #         idx = np.asarray(colors[cat], dtype=np.int64)
    #         scenes.append([ax.scatter(x_data[idx], y_data[idx], label=cat), names])
    # else:
    #     scenes.append([ax.scatter(x_data, y_data), names])

    # ax.grid()
    # ax.set_xscale("log" if _log_is_better(x_data) else "linear")
    # ax.set_yscale("log" if _log_is_better(y_data) else "linear")
    # ax.set_xlabel(kwargs["x"])
    # ax.set_ylabel(kwargs["y"])
    # plt.tight_layout()
    # if "color" in kwargs:
    #     plt.legend()

    # def update_annot(sc, ind, name):
    #     idx = ind["ind"][0]
    #     pos = sc.get_offsets()[idx]
    #     annot.xy = pos

    #     text = ""
    #     for key in ["compression_ratio", "crf", "encode_duration", "encode_threads", "encoder", "preset", "profile"]:
    #         text += f"{key}={data[name][key]}\n"
    #     text = text[:-1]
    #     annot.set_text(text)
    #     annot.get_bbox_patch().set_facecolor("lightyellow")
    #     annot.get_bbox_patch().set_alpha(0.8)

    #     # Calculer la position relative du point dans la figure
    #     x_disp, y_disp = ax.transData.transform(pos)
    #     width, height = fig.get_size_inches()*fig.dpi

    #     # Choisir un offset en fonction de l'espace autour du point
    #     offset_x = 20 if x_disp < width * 0.7 else -200  # vers la droite ou vers la gauche
    #     offset_y = 20 if y_disp < height * 0.7 else -100  # vers le haut ou le bas

    #     annot.set_position((offset_x, offset_y))

    # # Fonction appelée lors du mouvement de la souris
    # def hover(event):
    #     vis = annot.get_visible()
    #     if event.inaxes == ax:
    #         for i, (sc, names) in enumerate(scenes):
    #             cont, ind = sc.contains(event)
    #             if cont:
    #                 update_annot(sc, ind, names[i])
    #                 annot.set_visible(True)
    #                 fig.canvas.draw_idle()
    #                 break
    #             elif vis:
    #                 annot.set_visible(False)
    #                 fig.canvas.draw_idle()
    # annot = ax.annotate(
    #     "",
    #     xy=(0, 0),
    #     xytext=(10, 10),
    #     textcoords="offset points",
    #     bbox=dict(boxstyle="round", fc="w"),
    #     arrowprops=dict(arrowstyle="->"),
    # )
    # annot.set_visible(False)
    # fig.canvas.mpl_connect("motion_notify_event", hover)
    # plt.show()



#!/usr/bin/env python3

"""Initialize the SQL database."""

import pathlib
import sqlite3


ENV_UNIQUE = ["env_ffmpeg_version", "env_hostname", "env_logical_cores", "env_ram"]


def create_database(filename: str | bytes | pathlib.Path):
    """Create a new SQL database to store all video informations.

    Parameters
    ----------
    filename : pathlike
        The path of the new database to be created.

    Examples
    --------
    >>> import os, tempfile
    >>> from mendevi.database.create import create_database
    >>> create_database(database := tempfile.mktemp(suffix=".sqlite"))
    >>> os.remove(database)
    >>>
    """
    filename = pathlib.Path(filename).expanduser().resolve()
    assert not filename.exists(), f"the database has to be new, {filename} exists"

    with sqlite3.connect(filename) as sql_database:
        cursor = sql_database.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_act_activity (
            act_id INTEGER PRIMARY KEY AUTOINCREMENT,

            /* MEASURES */
            act_start TIMESTAMP NOT NULL,  -- absolute start timestamp
            act_duration FLOAT NOT NULL CHECK(act_duration > 0.0),  -- full encoding time in seconds
            act_rapl_dt LONGBLOB,  -- list of the duration of each interval in seconds
            act_rapl_power LONGBLOB,  -- list of the average power in watt in each interval
            act_wattmeter_dt LONGBLOB,  -- list of the duration of each interval in seconds
            act_wattmeter_power LONGBLOB,  -- list of the sampled power in watt in each point
            act_ps_dt LONGBLOB,  -- list of the duration of each interval in seconds
            act_ps_core LONGBLOB,  -- tensor of detailed usage of each logical core in %
            act_ps_ram LONGBLOB  -- list of the sampled ram usage in bytes in each point
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_vid_video (
            vid_id BINARY(128) PRIMARY KEY,  -- md5 hash of the video

            /* VIDEO CONTENT */
            vid_codec TINYTEXT,  -- the codec name
            vid_duration FLOAT CHECK(vid_duration > 0.0),  -- video duration in second
            vid_eotf TINYTEXT,  -- name of the electro optical transfer function
            vid_fps FLOAT CHECK(vid_fps > 0.0),  -- theorical framerate of the video
            vid_frames LONGBLOB,  -- serialized version of the metadata of all frames
            vid_gamut TINYTEXT,  -- name of the color space
            vid_height SMALLINT CHECK(vid_height > 0),  -- display height
            vid_size BIGINT CHECK(vid_width >= 0),  -- file size in bytes
            vid_subsampling TINYTEXT CHECK(vid_subsampling IN ('444', '422', '420')),  -- yuv pixel subsampling
            vid_width SMALLINT CHECK(vid_width > 0),  -- display width
            vid_bit_depth SMALLINT CHECK(vid_bit_depth IN (8, 10, 12)),  -- bit per sample

            /* METRICS */
            vid_lpips_alex LONGBLOB,  -- list lpips with alex for each frame
            vid_lpips_vgg LONGBLOB,  -- list lpips with vgg for each frame
            vid_psnr LONGBLOB,  -- list of the psnr (6, 1, 1) metric for each frame
            vid_ssim LONGBLOB,  -- list of the ssim (6, 1, 1) metric for each frame, gauss win 11x11
            vid_uvq LONGBLOB  -- list of the google uvq metric for each second of video
        )""")
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS t_env_environement (
            env_id INTEGER PRIMARY KEY AUTOINCREMENT,

            /* CONTEXT DETAILS */
            env_ffmpeg_version MEDIUMTEXT NOT NULL,
            env_hostname TINYTEXT NOT NULL,
            env_kernel_version TINYTEXT,
            env_libsvtav1_version MEDIUMTEXT,
            env_libvpx_vp9_version MEDIUMTEXT,
            env_libx265_version MEDIUMTEXT,
            env_logical_cores INTEGER NOT NULL CHECK(env_logical_cores > 0),
            env_lshw LONGTEXT,
            env_physical_cores INTEGER,
            env_pip_freeze MEDIUMTEXT,
            env_processor TINYTEXT,
            env_python_compiler TINYTEXT,
            env_python_version TINYTEXT,
            env_ram INTEGER NOT NULL CHECK(env_ram > 0),
            env_swap INTEGER,
            env_system_version MEDIUMTEXT,
            env_vvc_version MEDIUMTEXT,

            /* IDLE MEASURES */
            env_act_id INTEGER REFERENCES t_act_activity(act_id),  -- link to activity table

            /* CONSTRAINTS */
            UNIQUE({", ".join(ENV_UNIQUE)}) ON CONFLICT FAIL
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_dec_decode (
            dec_id INTEGER PRIMARY KEY AUTOINCREMENT,
            dec_vid_id BINARY(128) NOT NULL REFERENCES t_vid_video(vid_id) ON DELETE CASCADE,  -- link to video table
            dec_env_id INTEGER NOT NULL REFERENCES t_env_environement(env_id) ON DELETE CASCADE,  -- link to environement table
            dec_act_id INTEGER REFERENCES t_act_activity(act_id)  -- link to activity table
        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS t_enc_encode (
            enc_id INTEGER PRIMARY KEY AUTOINCREMENT,
            enc_vid_id BINARY(128) NOT NULL REFERENCES t_vid_video(vid_id) ON DELETE CASCADE,  -- link to video table
            enc_env_id INTEGER NOT NULL REFERENCES t_env_environement(env_id) ON DELETE CASCADE,  -- link to environement table
            enc_act_id INTEGER REFERENCES t_act_activity(act_id),  -- link to activity table

            /* TASK DESCRIPTION */
            enc_cmd TEXT,  -- exact ffmpeg command used
            enc_effort TINYTEXT CHECK(enc_effort IN ('fast', 'medium', 'slow')),  -- equivalent preset used for encoding
            enc_encoder TINYTEXT CHECK(enc_encoder IN ('libx264', 'libx265', 'libvpx-vp9', 'libsvtav1', 'vvc')),  -- the encoder name
            enc_file TEXT,  -- name of the source video file
            enc_quality FLOAT CHECK(enc_quality >= 0.0 AND enc_quality <= 1.0),  -- normlize crf in [0, 1]
            enc_threads SMALLINT CHECK(enc_threads >= 0)  -- number of threads used
        )""")
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_env_act_unicity
            BEFORE INSERT ON t_env_environement
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.env_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_update_env_act_unicity
            BEFORE UPDATE OF env_act_id ON t_env_environement
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.env_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_enc_act_unicity
            BEFORE INSERT ON t_enc_encode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.enc_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_enc_act_unicity
            BEFORE UPDATE OF enc_act_id ON t_enc_encode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.enc_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_dec_act_unicity
            BEFORE INSERT ON t_dec_decode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.dec_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.execute("""
            CREATE TRIGGER IF NOT EXISTS trigger_insert_dec_act_unicity
            BEFORE UPDATE OF dec_act_id ON t_dec_decode
            BEGIN
                SELECT
                    CASE
                        WHEN NEW.dec_act_id IN (
                            SELECT enc_act_id FROM t_enc_encode
                            UNION ALL SELECT dec_act_id FROM t_dec_decode
                            UNION ALL SELECT env_act_id FROM t_env_environement
                        ) THEN
                            RAISE (ABORT, 'act_id has to be unique')
                    END;
            END;
        """)
        cursor.close()


def is_sqlite(file: str | bytes | pathlib.Path):
    """Test if the provided path is an sqlite3 database.

    Examples
    --------
    >>> import os, pathlib, tempfile
    >>> from mendevi.database import create_database, is_sqlite
    >>> database = pathlib.Path(tempfile.mktemp())
    >>> is_sqlite(database)
    False
    >>> create_database(database)
    >>> is_sqlite(database)
    True
    >>> os.remove(database)
    >>>
    """
    file = pathlib.Path(file).expanduser().resolve()
    if not file.is_file():
        return False
    with open(file, "rb") as raw:
        header = raw.read(100)
    if len(header) < 100:  # SQLite database file header is 100 bytes
        return False
    return header.startswith(b"SQLite format 3")



#!/usr/bin/env python3

"""Scripts with enoslib for grid5000."""

import conf

LOCAL = False  # set True to execute the tests on the local computer

# local data
RES_LOCAL = "~/Téléchargements"
VIDEO_LOCAL = "/data/dataset/video/despacito.mp4"

# grid5000 data
RES_G5K = conf.DATA
VIDEO_G5K = f"{conf.DATA}/lossless_video/bbb.webm"

CMD = [
    "mendevi", "encode",
    "--profile", "sd", "--profile", "uhd",
    "--preset", "medium",
    "--encoder", "libx264", "--encoder", "libx265", "--encoder", "libsvtav1", "--encoder", "libvpx-vp9",
    "--points", "2",
    *(f"-t{i}" for i in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]),
    "--no-psnr", "--no-ssim", "--no-uvq", "--no-vmaf",
    "--out", f"{RES_LOCAL if LOCAL else RES_G5K}/encode_multithread_samples",
    "--res", f"{RES_LOCAL if LOCAL else RES_G5K}/encode_multithread.json",
    VIDEO_LOCAL if LOCAL else VIDEO_G5K,
]

if __name__ == "__main__":
    conf.run_script(CMD, LOCAL)



#!/usr/bin/env python3

"""Command line interface entry point."""

import click

from .draw import main as main_draw
from .encode import main as main_encode
from .merge import main as main_merge


@click.group()
def main() -> int:
    """Performs video transcoding measurements."""
    return 0


main.add_command(main_draw, "draw")
main.add_command(main_encode, "encode")
main.add_command(main_merge, "merge")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Provide simple tools."""

import logging
import numbers
import pathlib
import platform
import queue
import threading
import time
import typing

from cutcutcodec.core.io import VIDEO_SUFFIXES

from mendevi.g5kpower import g5kpower
from mendevi.rapl import RAPL
from mendevi.psutil import Usage


class Activity(threading.Thread):
    """Measure the computer activity of a section.

    Examples
    --------
    >>> import pprint, time
    >>> from mendevi.utils import Activity
    >>> with Activity() as activity:
    ...     time.sleep(1)
    ...
    >>> pprint.pprint(activity)
    >>>
    """

    def __init__(self, sleep: numbers.Real=50e-3):
        """Initalize the perf context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).
        """
        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep

        super().__init__(daemon=True)

        self._rapl_catcher = RAPL(sleep=sleep, no_fail=True)
        self._usage_catcher = Usage(sleep=sleep)
        self._exit_queue = queue.Queue()
        self.sleep = float(sleep)
        self.res: dict = {}

    def run(self):
        """Perform the measures."""
        self.res["start"] = time.time()
        with self._rapl_catcher as rapl, self._usage_catcher as usage:
            self._exit_queue.get()  # wait
        self.res |= {
            "ps_core": usage["cpu"],
            "ps_cores": usage["cpus"],
            "ps_dt": usage["dt"],
            "ps_ram": usage["ram"],
        }
        if rapl is not None:
            self.res |= {
                "rapl_dt": rapl["dt"],
                "rapl_energy": rapl["energy"],
                "rapl_power": rapl["power"],
                "rapl_powers": rapl["powers"],
            }

    def __enter__(self) -> dict:
        r"""Start to measure.

        Returns
        -------
        activity: dict[str]
            * duration: float, the real measure duration.
            * ps_core: float, the mean cummulated usage of all the logical cpus.
            * ps_cores: list[list[float]], tensor of detailed usage of each logical core in %.
            * ps_dt: list[float], the duration of each interval (in s).
            * ps_ram: list[int], list of the sampled ram usage in bytes in each point.
            * rapl_dt: list[float], the duration of each interval (in s).
            * rapl_energy: float, the total energy consumption (in J).
            * rapl_power: float, the average power, energy divided by the duration (in w).
            * rapl_powers: list[float], the average power in watt in each interval.
            * start: float, absolute timestamp.
            * wattmeter_dt: list[float],  the duration of each interval (in s).
            * wattmeter_energy: float, the total energy consumption (in J).
            * wattmeter_power: float, the average power, energy divided by the duration (in w).
            * wattmeter_powers: list[float], the sampled power in watt in each point.

        Notes
        -----
        The returned dictionary is update inplace when we exit the code bloc.
        Only the successfull field are created.
        """
        self.start()
        return self.res

    def __exit__(self, *_):
        """Stop the measure and update the dictionary returnd by __enter__."""
        # stop
        self.res["duration"] = time.time() - self.res["start"]
        self._exit_queue.put(None)
        self.join()
        # request wattmeter power
        try:
            wattmeter = g5kpower(platform.node(), self.res["start"], self.res["duration"])
        except ValueError:
            wattmeter = None
        else:
            self.res |= {
                "wattmeter_dt": wattmeter["dt"],
                "wattmeter_energy": wattmeter["energy"],
                "wattmeter_power": wattmeter["power"],
                "wattmeter_power": wattmeter["powers"],
            }


def unfold_video_files(
    paths: typing.Iterable[str | bytes | pathlib.Path]
) -> typing.Iterable[pathlib.Path]:
    """Explore recursively the folders to find the video path.

    Parameters
    ----------
    paths : list[pathlike]
        All the folders, files, glob or recursive glob expression.

    Yields
    ------
    filename : pathlib.Path
        The full path of the image.
    """
    assert hasattr(paths, "__iter__"), paths.__class__.__name__
    for path in paths:
        path = pathlib.Path(path).expanduser().resolve()
        if path.is_file():
            yield path
        elif path.is_dir():
            for root, _, files in path.walk():
                for file in files:
                    file = root / file
                    if file.suffix.lower() in VIDEO_SUFFIXES:
                        yield file
        elif "*" in path.name and path.parent.is_dir():
            yield from unfold_video_files(path.parent.glob(path.name))
        elif "**" in (parts := path.parts):
            idx = parts.index("**")
            yield from unfold_video_files(
                pathlib.Path(*parts[:idx]).glob(pathlib.Path(*parts[idx:]))
            )
        else:
            logging.warning("the path %s is not correct", path)



#!/usr/bin/env python3

"""Get the information of the environement."""

import logging
import os
import platform
import re
import subprocess

import psutil


def get_ffmpeg_version() -> str:
    r"""Return the version of ffmpeg.

    Examples
    --------
    >>> import pprint
    >>> from mendevi.context import get_ffmpeg_version
    >>> pprint.pprint(get_ffmpeg_version())
    ('ffmpeg version N-120220-g934e1c23b0 Copyright (c) 2000-2025 the FFmpeg '
     'developers\n'
     'built with gcc 14 (Ubuntu 14.2.0-4ubuntu2)\n'
     'configuration: --prefix=/home/rrichard/ffmpeg_build '
     '--pkg-config-flags=--static '
     '--extra-cflags=-I/home/rrichard/ffmpeg_build/include '
     "--extra-ldflags=-L/home/rrichard/ffmpeg_build/lib --extra-libs='-lpthread "
     "-lm' --ld=g++ --bindir=/home/rrichard/bin --enable-gpl --enable-gnutls "
     '--enable-libaom --enable-libass --enable-libfdk-aac --enable-libfreetype '
     '--enable-libmp3lame --enable-libopus --enable-libsvtav1 --enable-libdav1d '
     '--enable-libvorbis --enable-libvpx --enable-libx264 --enable-libx265 '
     '--enable-libvvenc --enable-nonfree\n'
     'libavutil      60.  4.101 / 60.  4.101\n'
     'libavcodec     62.  6.100 / 62.  6.100\n'
     'libavformat    62.  1.101 / 62.  1.101\n'
     'libavdevice    62.  0.100 / 62.  0.100\n'
     'libavfilter    11.  1.100 / 11.  1.100\n'
     'libswscale      9.  0.100 /  9.  0.100\n'
     'libswresample   6.  0.100 /  6.  0.100')
    """
    out = subprocess.run(
        ["ffmpeg", "-version"], check=False, capture_output=True
    ).stdout.decode()
    lines = out.split("\n")
    lines = [l for l in lines if l not in {"", "Exiting with exit code 0"}]
    out = "\n".join(lines)
    return out


def get_libx265_version() -> str:
    """Return the version of the libx265 encoder.

    Examples
    --------
    >>> from mendevi.context import get_libx265_version
    >>> print(get_libx265_version())
    4.1+188-cd4f0d6e9
    [Linux][GCC 14.2.0][64 bit] 8bit+10bit+12bit
    MMX2 SSE2Fast LZCNT SSSE3 SSE4.2 AVX FMA3 BMI2 AVX2
    >>>
    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libx265",
            "-f", "null", os.devnull,
        ],
        check=False, capture_output=True
    ).stderr.decode()

    # version
    if (match := re.search(r"HEVC encoder version (?P<version>\S+)", out)) is None:
        logging.warning("failed to find the libx265 version")
        version = ""
    else:
        version = match["version"]
    # build
    if (match := re.search(r"build info (?P<build>.+)\n", out)) is None:
        logging.warning("failed to find the libx265 build info")
        build = ""
    else:
        build = match["build"]
    # cpu
    if (match := re.search(r"using cpu capabilities: (?P<cpu>.+)\n", out)) is None:
        logging.warning("failed to find the libx265 build info")
        cpu = ""
    else:
        cpu = match["cpu"]

    out = f"{version}\n{build}\n{cpu}"
    return out


def get_libvpx_vp9_version() -> str:
    """Return the version of the libx265 encoder.

    Examples
    --------
    >>> from mendevi.context import get_libvpx_vp9_version
    >>> print(get_libvpx_vp9_version())
    v1.14.1
    >>>
    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libvpx-vp9",
            "-f", "null", os.devnull,
        ],
        check=False, capture_output=True
    ).stderr.decode()
    if (match := re.search(r"\[libvpx-vp9 @ \w+\] (?P<version>\S+)", out)) is None:
        logging.warning("failed to find the libvpx-vp9 version")
        return ""
    return match["version"]


def get_libsvtav1_version() -> str:
    """Return the version of the libx265 encoder.

    Examples
    --------
    >>> from mendevi.context import get_libsvtav1_version
    >>> print(get_libsvtav1_version())
    v3.0.2-103-gc10214fa
    GCC 14.2.0 64 bit
    >>>
    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "libsvtav1",
            "-f", "null", os.devnull,
        ],
        check=False, capture_output=True
    ).stderr.decode()

    # version
    if (match := re.search(r"SVT-AV1 Encoder Lib (?P<version>\S+)", out)) is None:
        logging.warning("failed to find the libsvtav1 version")
        version = ""
    else:
        version = match["version"]
    # build
    if (match := re.search(r"SVT \[build\]\s*:\s*(?P<build>\S.*\S)\s*\n", out)) is None:
        logging.warning("failed to find the libsvtav1 build info")
        build = ""
    else:
        build = re.sub(r"\s+", " ", match["build"])

    out = f"{version}\n{build}"
    return out


def get_vvc_version() -> str:
    """Return the version of the libx265 encoder.

    Examples
    --------
    >>> from mendevi.context import get_vvc_version
    >>> print(get_vvc_version())
    1.13.1
    >>>
    """
    out = subprocess.run(
        [
            "ffmpeg", "-f", "lavfi", "-i", "nullsrc", "-frames:v", "1",
            "-c:v", "vvc",
            "-f", "null", os.devnull,
        ],
        check=False, capture_output=True
    ).stderr.decode()
    if (match := re.search(r"libvvenc version: (?P<version>\S+)", out)) is None:
        logging.warning("failed to find the vvc version")
        return ""
    return match["version"]


def get_platform() -> dict:
    """Get basic information about the system.

    Examples
    --------
    >>> import pprint
    >>> from mendevi.context import get_platform
    >>> pprint.pprint(get_platform())
    {'hostname': 'PTB-5CG4414JJD',
     'kernel_version': 'Linux-6.11.0-29-generic-x86_64-with-glibc2.40',
     'processor': 'x86_64',
     'python_compiler': 'x86_64',
     'python_version': 'x86_64',
     'system_version': '#29-Ubuntu SMP PREEMPT_DYNAMIC Fri Jun 13 20:29:41 UTC '
                       '2025'}
    >>>
    """
    return {
        "hostname": platform.node(),
        "kernel_version": platform.platform(),
        "logical_cores": psutil.cpu_count(logical=True),
        "physical_cores": psutil.cpu_count(logical=False),
        "processor": platform.processor(),
        "python_compiler": platform.python_compiler(),
        "python_version": platform.python_version(),
        "ram": psutil.virtual_memory().total,
        "swap": psutil.swap_memory().total,
        "system_version": platform.version(),
    }


def get_pip_freeze() -> str:
    """Return the sorted pip freeze.

    Examples
    --------
    >>> from mendevi.context import get_pip_freeze
    >>> print(get_pip_freeze())
    """
    out = subprocess.run(
        ["pip", "freeze"], check=False, capture_output=True,
    ).stdout.decode()
    return out


def get_lshw() -> str:
    """The very accurate exaustive info as big json.

    Examples
    --------
    >>> from mendevi.context import get_lshw
    >>> full_info = get_lshw()
    >>>
    """
    if os.geteuid() != 0:
        logging.warning("you should run as super user to get more system info")
    try:
        out = subprocess.run(
            ["lshw", "-json"], check=False, capture_output=True,
        ).stdout.decode()
    except FileNotFoundError:
        logging.error("please install lshw: sudo apt install lshw")
    return out


def full_context() -> dict:
    """Get the full context informations."""
    return {
        "ffmpeg_version": get_ffmpeg_version(),
        "libx265_version": get_libx265_version(),
        "libvpx_vp9_version": get_libvpx_vp9_version(),
        "libsvtav1_version": get_libsvtav1_version(),
        "vvc_version": get_vvc_version(),
        "pip_freeze": get_pip_freeze(),
        "lshw": get_lshw(),
        **get_platform()
    }



#!/usr/bin/env python3

"""Perform encoding measures."""

import fractions
import hashlib
import itertools
import pathlib
import re
import shlex
import shutil
import sqlite3
import subprocess
import tempfile

from context_verbose import Printer
import click
import cutcutcodec
import numpy as np
import tqdm

from mendevi.database.complete import add_environement
from mendevi.database.create import create_database, is_sqlite
from mendevi.database.serialize import list_to_binary, tensor_to_binary
from .utils import Activity
from .profiles import PROFILES


ENCODERS = {"libx264", "libx265", "libvpx-vp9", "libsvtav1", "vvc"}


def encode(src: pathlib.Path, *args, **kwargs) -> tuple[pathlib.Path, dict[str]]:
    """Transcode an existing video.

    Parameters
    ----------
    src : pathlib.Path
        The source video file to be transcoded.
    *args, **kwargs
        Transmitted to :py:func:`get_transcode_cmd`.

    Returns
    -------
    dst : pathlib.Path
        The transcoded video path. The stem contains the md5 hash of the file content.
    cmd : str
        The ffmpeg command
    activity : dict[str]
        The computeur activity during the transcoding process.
    """
    assert isinstance(src, pathlib.Path), src.__class__.__name__
    assert src.is_file(), src

    # find tempfile name
    cmd = get_transcode_cmd(*args, **kwargs)
    signature = hashlib.md5(bytes(src) + (" ".join(cmd)).encode("utf-8")).hexdigest()
    dst = pathlib.Path(tempfile.gettempdir()) / f"{signature}.mp4"

    # display
    with Printer(f"Execute: <{' '.join(map(shlex.quote, cmd))}>...", color="white") as prt:
        prt.print(f"input video: {src.name}")
        duration = round(float(cutcutcodec.get_duration_video(src)), 2)
        load = tqdm.tqdm(
            dynamic_ncols=True,
            leave=False,
            smoothing=1e-2,
            total=duration,
            unit="s",
        )

        # transcode
        cmd = [
            "ffmpeg", "-y", "-hide_banner",
            "-i", str(src), "-c:v", *cmd,
            str(dst),
        ]
        with Activity() as activity, subprocess.Popen(cmd, stderr=subprocess.PIPE) as process:
            out = b""
            is_finish = False
            while not is_finish:
                while (
                    match := re.search(br"time=(?P<h>\d+):(?P<m>\d{1,2}):(?P<s>\d{1,2}\.\d*)", out)
                ) is None:
                    if not (buff := process.stderr.read(32)):
                        is_finish = True
                        break
                    out += buff
                else:
                    out = out[match.endpos:]
                    elapsed = round(
                        3600.0*float(match["h"]) + 60.0*float(match["m"]) + float(match["s"]),
                        2,
                    )
                    load.total = max(load.total, elapsed)
                    load.update(elapsed-load.n)
            load.close()
            if process.returncode:
                raise RuntimeError(f"failed to execute {cmd}")

        # print
        prt.print(f"avg cpu usage: {activity['ps_core']:.1f} %")
        prt.print(f"avg ram usage: {1e-9*np.mean(activity['ps_ram']):.2g} Go")
        if "rapl_power" in activity:
            prt.print(f"avg rapl power: {activity['rapl_power']:.2g} W")
        if "wattmeter_power" in activity:
            prt.print(f"avg wattmeter power: {activity['wattmeter_power']:.2g} W")

        # compute file hash
        with open(dst, "rb") as raw:
            signature = hashlib.file_digest(raw, "md5").hexdigest()
        prt.print(f"output video: sample_{signature}.mp4")

    # move file
    final_dst = src.parent / f"sample_{signature}.mp4"
    if not final_dst.exists():
        shutil.copy(dst, src.parent / f"sample_{signature}_partial.mp4")
        shutil.move(src.parent / f"sample_{signature}_partial.mp4", final_dst)
    dst.unlink()

    cmd = " ".join(map(shlex.quote, cmd))
    return final_dst, cmd, activity


def encode_and_store(
    database: sqlite3.Cursor | str | bytes | pathlib.Path,
    env_id: int,
    src: pathlib.Path,
    **kwargs,
):
    """Transcode a video file and store the result in the database.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    **kwargs
        Transmitted to :py:func:`encode`.

    Examples
    --------
    >>> import pathlib, tempfile
    >>> from mendevi.database.complete import add_environement
    >>> from mendevi.database.create import create_database
    >>> from mendevi.encode import encode_and_store
    >>> src = pathlib.Path("/data/dataset/video/despacito.mp4")
    >>> create_database(database := pathlib.Path(tempfile.mktemp(suffix=".sqlite")))
    >>> env_id = add_environement(database)
    >>> encode_and_store(
    ...     database, env_id, src, encoder="libx264", profile="sd", effort="fast", quality=0.5, threads=8
    ... )
    >>> database.unlink()
    >>>
    """
    # open database
    if not isinstance(database, sqlite3.Cursor):
        database = pathlib.Path(database).expanduser().resolve()
        assert is_sqlite(database), database
        with sqlite3.connect(database) as sql_database:
            # sql_database.execute("PRAGMA journal_mode=WAL")
            cursor = sql_database.cursor()
            res = encode_and_store(cursor, env_id, src, **kwargs)
            cursor.close()
            return res

    # transcode the video
    file, cmd, activity = encode(src, **kwargs)

    # fill video table
    vid_id = re.search(r"[0-9a-f]{32}", file.stem).group()
    try:
        database.execute("INSERT INTO t_vid_video (vid_id) VALUES (?)", (vid_id,))
    except sqlite3.IntegrityError:
        pass

    # fill activity table
    activity = {
        "act_duration": activity["duration"],
        "act_ps_core": tensor_to_binary(activity["ps_cores"]),
        "act_ps_dt": list_to_binary(activity["ps_dt"]),
        "act_ps_ram": list_to_binary(activity["ps_ram"]),
        "act_rapl_dt": list_to_binary(activity.get("rapl_dt", None)),
        "act_rapl_power": list_to_binary(activity.get("rapl_powers", None)),
        "act_start": activity["start"],
        "act_wattmeter_dt": list_to_binary(activity.get("wattmeter_dt", None)),
        "act_wattmeter_power": list_to_binary(activity.get("wattmeter_powers", None)),
    }
    keys = list(activity)
    (act_id,) = database.execute(
        (
            f"INSERT INTO t_act_activity ({', '.join(keys)}) "
            f"VALUES ({', '.join('?'*len(keys))}) RETURNING act_id"
        ),
        [activity[k] for k in keys],
    ).fetchone()

    # fill encode table
    database.execute("""INSERT INTO t_enc_encode
        (
            enc_vid_id, enc_env_id, enc_act_id,
            enc_cmd, enc_effort, enc_encoder, enc_file, enc_quality, enc_threads
        )
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)""",
        (
            vid_id, env_id, act_id,
            cmd, kwargs["effort"], kwargs["encoder"], src.name, kwargs["quality"], kwargs["threads"]
        )
    )


def get_transcode_cmd(
    encoder: str, effort: str, quality: float, threads: int
) -> list[str]:
    """Return the ffmpeg encode cmd."""
    assert isinstance(encoder, str), encoder.__class__.__name__
    assert encoder in ENCODERS, encoder
    assert isinstance(effort, str), effort.__class__.__name__
    assert isinstance(quality, float), quality.__class__.__name__
    assert isinstance(threads, int), threads.__class__.__name__
    assert effort in {"slow", "medium", "fast"}, effort
    assert 0.0 <= quality <= 1.0, quality
    assert threads > 0, threads

    def libsvtav1_lp(threads: int) -> int:
        """Convert threads in parralel level."""
        # https://gitlab.com/AOMediaCodec/SVT-AV1/-/blob/master/Source/Lib/Globals/enc_handle.c#L598
        # lp=1 -> threads=1
        # lp=2 -> threads=2
        # lp=3 -> threads=8
        # lp=4 -> threads=12
        # lp=5 -> threads=16
        # lp=6 -> threads=20
        return {  # threads to lp
            1: 1,
            2: 2, 3: 2, 4: 2, 5: 2,
            6: 3, 7: 3, 8: 3, 9: 3, 10: 3,
            11: 4, 12: 4, 13: 4, 14: 4,
            15: 5, 16: 5, 17: 5,
        }.get(threads, 6)

    match encoder:
        case "libx264":
            return [  # https://ffmpeg.party/x264/
                "libx264",
                "-crf", str(round(quality*51.0, 1)),
                "-preset", effort,
                "-tune", "ssim",
                "-threads", str(threads),
            ]
        case "libx265":
            return [   # https://x265.readthedocs.io/en/master/cli.html
                "libx265",
                "-crf", str(round(quality*51.0, 1)),
                "-preset", effort,
                "-tune", "ssim",
                "-x265-params",
                f"frame-threads={threads}:pools={threads}:wpp={1 if threads != 1 else 0}",
            ]
        case "libvpx-vp9":
            return [  # https://wiki.webmproject.org/ffmpeg/vp9-encoding-guide
                "libvpx-vp9",
                "-crf", str(round(quality*63.0)),
                "-speed", {"slow": "-2", "medium": "1", "fast": "8"}[effort],  # in [-16, 16]
                "-tune", "ssim",
                "-row-mt", "1", "-threads", str(threads),
            ]
        case "libsvtav1":
            return [
                "libsvtav1",
                "-crf", str(round(quality*63.0)),
                "-preset", {"slow": "4", "medium": "6", "fast": "8"}[effort],
                "-svtav1-params", f"film-grain=0:lp={libsvtav1_lp(threads)}",  # tune=2 broken in v3.0.2
                # "-threads", str(threads),
            ]
        case "vvc":
            profile = "uhd"  # TODO: détecter automatiquement grace a la video source
            bit = int(re.search(r"(?P<bit>\d+)le", PROFILES[profile]["pix_fmt"] + "8le")["bit"])
            return [
                "vvc",
                "-qp", str(round(quality*63.0)),
                "-preset", effort,
                "-qpa", "1",
                "-vvenc-params", f"internalbitdepth={bit}",
                "-threads", str(threads),
            ]


@click.command()
@click.argument("videos", type=click.Path(), nargs=-1)
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--repeat",
    type=int,
    default=2,
    help="The number of times the experiment is repeated.",
)
@click.option(
    "-e", "--effort",
    type=click.Choice(["fast", "medium", "slow"]),
    default=["medium"],
    multiple=True,
    help="The compression effort (default = medium).",
)
@click.option(
    "-c", "--encoder",
    type=click.Choice(sorted(ENCODERS)),
    default=sorted(ENCODERS),
    multiple=True,
    help="The encoder name.",
)
@click.option(
    "-n", "--points",
    type=int,
    default=24,
    help="The number of quality point per encoder.",
)
@click.option(
    "-t", "--threads",
    type=int,
    default=[8],
    multiple=True,
    help="The number of threads used by encoders.",
)
def main(videos: tuple, database: str = None, **kwargs):
    """Measures activity during encoding.

    \b
    Parameters
    ----------
    videos : tuple[pathlike]
        The source videos to be transcoded.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        By default, it is created right next to the video.
    repeat : int, default=2
        The number of times the experiment is repeated.
        This allows us to estimate the variance of measurements.
    effort : tuple[str], default=("medium",)
        The effort made to compress, `fast`, `medium` or `slow`.
    encoder : tuple[str], default=("libsvtav1", "libvpx-vp9", "libx264", "libx265", "vvc")
        The encoders and therefore the codecs to use.
    points : int, default=24
        The number of different qualities to use.
        It is an indirect way to determine the CRF or the QP.
        The quality values are distributed evenly over ]0, 1[,
        for example, points=3 => qualities=[0.25, 0.5, 0.75].
    threads : int, default=8
        The theoretical number of threads used by the encoder.
        This roughly reflects the number of logical cores used.
    """
    with Printer("Parse configuration...") as prt:
        # video
        assert isinstance(videos, tuple), videos.__class__.__name__
        assert videos, "at least one video is required"
        videos = [pathlib.Path(v).expanduser() for v in videos]
        assert all(v.is_file() for v in videos), videos
        prt.print(f"videos   : {', '.join(map(str, videos))}")

        # database
        database = pathlib.Path(database or videos[0].parent / "mendevi.db").expanduser()
        if not database.exists():
            create_database(database)
        prt.print(f"database : {database}")

        # repeat
        assert "repeat" in kwargs, sorted(kwargs)
        assert isinstance(kwargs["repeat"], int), kwargs["repeat"].__class__.__name__
        assert kwargs["repeat"] >= 1, kwargs["repeat"]
        prt.print(f"repeat   : {kwargs['repeat']}")

        # effort
        assert "effort" in kwargs, sorted(kwargs)
        assert isinstance(kwargs["effort"], tuple), kwargs["effort"].__class__.__name__
        assert all(isinstance(p, str) for p in kwargs["effort"]), kwargs["effort"].__class__.__name__
        assert all(p in {"fast", "medium", "slow"} for p in kwargs["effort"]), kwargs["effort"]
        prt.print(f"efforts  : {', '.join(kwargs['effort'])}")

        # encoder
        assert "encoder" in kwargs, sorted(kwargs)
        assert isinstance(kwargs["encoder"], tuple), kwargs["encoder"].__class__.__name__
        assert all(isinstance(e, str) for e in kwargs["encoder"]), kwargs["encoder"].__class__.__name__
        assert all(e in ENCODERS for e in kwargs["encoder"]), kwargs["encoder"]
        prt.print(f"encoders : {', '.join(kwargs['encoder'])}")

        # points
        assert "points" in kwargs, sorted(kwargs)
        assert isinstance(kwargs["points"], int), kwargs["points"].__class__.__name__
        assert kwargs["points"] >= 1, kwargs["points"]
        kwargs["quality"] = [
            fractions.Fraction(i+1, kwargs["points"]+2)
            for i in range(kwargs["points"])
        ]
        prt.print(f"qualities: k * {kwargs['quality'][0]}, k \u2208 [1, {kwargs['points']}]")
        kwargs["quality"] = list(map(float, kwargs["quality"]))

        # threads
        assert "threads" in kwargs, sorted(kwargs)
        assert isinstance(kwargs["threads"], tuple), kwargs["threads"].__class__.__name__
        assert all(isinstance(t, int) for t in kwargs["threads"]), kwargs["threads"].__class__.__name__
        assert all(t >= 1 for t in kwargs["threads"]), kwargs["threads"]
        prt.print(f"threads  : {', '.join(map(str, kwargs['threads']))}")

    # retrieves the settings for videos that have already been transcoded
    env_id = add_environement(database)
    with sqlite3.connect(database) as sql_database:
        # sql_database.execute("PRAGMA journal_mode=WAL")
        done: dict[tuple, int] = {}
        for key in sql_database.execute(
            """
            SELECT enc_effort, enc_encoder, enc_file, enc_quality, enc_threads
            FROM t_enc_encode WHERE enc_env_id=?
            """,
            (env_id,)
        ):
            done[key] = done.get(key, 0) + 1

    # iterate on all the parameters
    nbr = (
        kwargs["repeat"]
        * len(kwargs["threads"])
        * len(kwargs["effort"])
        * kwargs["points"]
        * len(kwargs["encoder"])
    )
    for i, (repeat, video, threads, effort, quality, encoder) in enumerate(
        itertools.product(
            range(kwargs["repeat"]),
            videos,
            kwargs["threads"],
            kwargs["effort"],
            kwargs["quality"],
            kwargs["encoder"],
        )
    ):
        key = (effort, encoder, video.name, quality, threads)
        if done.get(key, 0) > repeat:
            continue
        with Printer(
            (
                f"Encode {i}/{nbr}: "
                f"video={video.name}, "
                f"threads={threads}, "
                f"effort={effort}, "
                f"quality={quality:.2f}, "
                f"encoder={encoder}..."
            ),
            color="cyan",
        ) as prt:
            encode_and_store(
                database, env_id, video,
                threads=threads, effort=effort, quality=quality, encoder=encoder,
            )
            done[key] = done.get(key, 0) + 1
            prt.print_time()



#!/usr/bin/env python3

"""Merges multiple json contents."""

import json
import pathlib
import shutil

import click
import filelock


def merge_data(data_ref: object, data_new: object) -> object:
    """Merge the new data into the ref data.

    Parameters
    ----------
    data_ref : object
        The jsonisable object that will be extended by data_new.
        It can be modified inplace.
    data_new : object
        The jsonisable object to be added to the reference content.

    Returns
    -------
    data_merge : object
        Merging data_ref and data_new. Can be a data_ref reference.

    Examples
    --------
    >>> from mendevi.merge import merge_data
    >>> merge_data("text 1", "text 2")
    ['text 1', 'text 2']
    >>> merge_data("text 1", "text 2")
    >>> merge_data(["text 1"], "text 2")
    ['text 1', 'text 2']
    >>> merge_data({"key1": "d1_1", "key2": "d1_2"}, {"key2": "d2_2", "key3": "d2_3"})
    {'key1': 'd1_1', 'key2': ['d1_2', 'd2_2'], 'key3': 'd2_3'}
    >>>
    """
    def list_depth(item: object, _depth: int = 0) -> int:
        """Return the number of neasted lists."""
        if isinstance(item, list):
            return min(list_depth(e, _depth+1) for e in item)
        return _depth

    if list_depth(data_ref) == list_depth(data_new) + 1:
        data_ref.append(data_new)
        return data_ref
    if isinstance(data_ref, dict) and isinstance(data_new, dict):
        for new_key, new_value in data_new.items():
            if new_key not in data_ref:
                data_ref[new_key] = new_value
            else:
                data_ref[new_key] = merge_data(data_ref[new_key], new_value)
        return data_ref
    return [data_ref, data_new]  # default case


def merge_file(file: pathlib.Path | str | bytes, data_new: object):
    """Add a content into a json file.

    Parameters
    ----------
    file : pathlike
        A json file to be updated with new content.
    data_new : object
        The jsonisable object to be added to the json file.

    Notes
    -----
    This function is locked so that it can be called from different processes.
    """
    file = pathlib.Path(file).expanduser()

    with filelock.FileLock(file.with_suffix(".lock")):
        try:
            with open(file, "r", encoding="utf-8") as raw:
                data_ref = json.load(raw)
        except FileNotFoundError:
            data_ref = {}
        data_ref = merge_data(data_ref, data_new)
        with open(file.with_suffix(".tmp"), "w", encoding="utf-8") as raw:
            json.dump(data_ref, raw, sort_keys=True, indent=4)
        shutil.move(file.with_suffix(".tmp"), file)


@click.command()
@click.argument("files", nargs=-1, type=click.Path())
@click.option("-r", "--res", type=click.Path(), help="The result merged json file.")
def main(files: list, res: str):
    """Merge several files."""
    # preparation
    assert len(files) >= 2, f"You must provide at least 2 files to be merged."
    files = [pathlib.Path(f).expanduser() for f in files]
    assert all(f.is_file() for f in files), f"One of the files {files} doese not exists."
    res = pathlib.Path(res or pathlib.Path.cwd() / "merged.json").expanduser()

    # merge
    data = {}
    for file in files:
        with open(file, "r", encoding="utf-8") as raw:
            data_inter = json.load(raw)
        data = merge_data(data, data_inter)

    # write
    res.unlink(missing_ok=True)
    merge_file(res, data)



#!/usr/bin/env python3

"""Set the grid5000 configuration variables.

If you are using pyenv on grid5000, you must help ansible to find the right python interpreter.
In the file ``~/.ansible.cfg``, put the following content:
[defaults]
interpreter_python = /root/.pyenv/versions/3.13.3/envs/3.13/bin/python
"""

import json
import shlex
import subprocess
import time

import enoslib

CLUSTER = "paradoxe"
SERVERS = ["paradoxe-7.rennes.grid5000.fr"]  # optional
DATA = "/srv/storage/videoimpact@storage3.rennes.grid5000.fr"
ENV_NAME = "https://api.grid5000.fr/sid/sites/rennes/public/rrichard/mendevi_env.yaml"
ROLES = ["magelan"]
WALLTIME = "72:00:00"

SLEEP_INIT = 10.0  # initial sleeping time
SLEEP_FINAL = 600.0  # final sleeping time
SLEEP_TRANSITION = 3600.0  # the transition time between the init and final sleep


def run_script(cmd: str | list[str], local: bool=False):
    """Run a script on the grid."""
    # parse the argument
    assert isinstance(local, bool), local.__class__.__name__
    if isinstance(cmd, list):
        assert all(isinstance(e, str) for e in cmd), cmd
        cmd = " ".join(map(shlex.quote, cmd))
    else:
        assert isinstance(cmd, str), cmd.__class__.__name__

    # run localy
    if local:
        subprocess.run(cmd, shell=True)
        return

    # run on the grid
    enoslib.init_logging()
    conf = (  # doc: https://discovery.gitlabpages.inria.fr/enoslib/apidoc/infra.html#g5k-schema
        enoslib.G5kConf.from_settings(
            env_name=ENV_NAME,
            job_name="Mesures Encodage et Decodage Video",
            job_type=["deploy"],
            key="~/.ssh/id_ed25519.pub",
            walltime=WALLTIME,
            monitor="wattmetre_power_watt",
            # queue="production",  # "default", "testing", "besteffort"
        )
        .add_machine(roles=ROLES, cluster=CLUSTER, nodes=1, **({"servers": SERVERS} if SERVERS else {}))
    )
    sleep = SLEEP_INIT
    with enoslib.G5k(conf) as (roles, _):  # get and release all grid5000 resources
        main_results = enoslib.run_command(cmd, roles=roles, background=True)
        log_files = log_files = {result.results_file for result in main_results}
        while log_files:
            # waiting a moment
            # at the beginning, we wait SLEEP_INIT, and at the end, we wait SLEEP_FINAL
            # it takes n steps, with n = math.log(SLEEP_FINAL/SLEEP_INIT) / math.log(alpha)
            # with alpha = 1.0 + (SLEEP_FINAL - SLEEP_INIT) / SLEEP_TRANSITION
            # the full transition time take SLEEP_TRANSITION
            time.sleep(sleep)
            sleep *= 1.0 + (SLEEP_FINAL - SLEEP_INIT) / SLEEP_TRANSITION
            sleep = min(sleep, SLEEP_FINAL)
            for log_file in log_files.copy():
                for res in enoslib.run_command(f"cat {log_file}", roles=roles, on_error_continue=True):
                    res = json.loads(res.payload["stdout"])
                    if (msg := res.get("stderr", "")):
                        raise RuntimeError(msg)
                    elif "rc" in res:
                        print(res["stdout"])
                        log_files.remove(log_file)



#!/usr/bin/env python3

"""Test the wattmeter on grid5000."""

import conf

CMD = [
    "python",
    "-c",
    (
        "import platform, time; from mendevi.g5kpower import g5kpower; "
        "print(g5kpower(platform.node(), time.time()-10.0, 5.0))"
    )
]

if __name__ == "__main__":
    conf.run_script(CMD, local=False)



#!/usr/bin/env python3

"""Structure of a scalar to represent the value of a physical quantity.

Defines a scalar physical quantity to avoid omogeneity errors by being very clear about the units,
then helps propagate uncertainties.

The motivation is to help to be clear on the normalisation rules.
"""

import math
import numbers
import re
import typing

import sympy


class Scalar:
    """A constant scalar to represent the value of a physical quantity.

    Attributes
    ----------
    avg : float
        The value of itself (readonly).
    std : float
        The standard deviation (readonly).

    Examples
    --------
    >>> from mendevi.scalar import Scalar
    >>> (nbr := Scalar(1/3, std=0.01))
    Scalar(0.3333333333333333, std=0.01)
    >>> print(nbr)
    0.33 \xb1 0.03
    >>> nbr.std
    0.01
    >>>
    >>> Scalar(nbr)  # copy data
    Scalar(0.3333333333333333, std=0.01)
    >>>
    """

    def __init__(
        self,
        avg: numbers.Real | str | typing.Self,
        *,
        std: numbers.Real | str=0,
    ):
        """Initialise the attributes."""
        if isinstance(avg, self.__class__):  # from a Scalar
            self._avg, self._std, self._unit = avg._avg, avg._std, avg._unit
        else:
            self._avg = float(avg)
            self._std = float(std)
            assert self._std >= 0, f"the standard deviation ({std}) has to be >=0"

    def __add__(self, other: numbers.Real | str | typing.Self) -> typing.Self:
        """Add self and other, assuming they are following 2 normal independant laws."""
        try:
            other = Scalar(other)
        except ValueError:
            return NotImplemented
        avg = self._avg + other._avg  # E[X1+X2] = E[x1]+E[X2]
        std = math.sqrt(self._std**2 + other._std**2)
        return self.__class__(avg, std=std)

    def __eq__(self, other: typing.Self) -> bool:
        """For hash table."""
        return self._avg == Scalar(other).avg

    def __float__(self) -> float:
        """Convert into single float."""
        return self._avg

    def __hash__(self) -> int:
        """Imuable hash table behavor."""
        return hash((self._avg, self._std))

    def __mul__(self, other: numbers.Real | str | typing.Self) -> typing.Self:
        """Multiply self and other, assuming they are following 2 normal independant laws.

        Examples
        --------
        >>> from lca.classes.scalar import Scalar
        >>> Scalar(2, std=1/2) * Scalar(4, std=1/2)
        Scalar(8.0, std=2.25)
        >>>
        """
        try:
            other = Scalar(other)
        except ValueError:
            return NotImplemented
        avg = self._avg * other._avg  # E[X1.X2] = E[X1].E[X2]
        var1, var2 = self._std * self._std, other._std * other._std
        mu12, mu22 = self._avg * self._avg, other._avg * other._avg
        std = math.sqrt(var1*var2 + var1*mu22 + var2*mu12)  # V[X1.X2] = E[X1.X1.X2.X2]-E[X1.X2]**2
        return self.__class__(avg, std=std)

    def __radd__(self, other: numbers.Real | str | typing.Self) -> typing.Self:
        """Addition is commutative."""
        return self.__add__(other)

    def __repr__(self) -> str:
        """Gives a nice evaluable representation of itself."""
        args = [str(self._avg)]
        if self._std != 0:
            args.append(f"std={self._std}")
        return f"{self.__class__.__name__}({', '.join(args)})"

    def __rmul__(self, other: numbers.Real | str | typing.Self) -> typing.Self:
        """Product is commutative."""
        return self.__mul__(other)

    def __str__(self) -> str:
        """Gives a nice compact representation of itself."""
        out = f"{float(self):.2g}"
        if self._std != 0:
            out += f" \xb1 {3*self._std:.2g}"  # utf8 "b1" is +- symbol
        if self._unit != 1:
            out += f" {self._unit}"
        return out

    @property
    def avg(self) -> float:
        """The representative value as a float number."""
        return self._avg

    @property
    def std(self) -> float:
        """The standard deviation."""
        return self._std



#!/usr/bin/env python3

"""Perform encoding measures."""

import itertools
import pathlib
import platform
import re
import shutil
import subprocess
import tempfile
import threading
import time

import click
import cutcutcodec
import numpy as np
import tqdm

from .g5kpower import G5kPower, g5kpower
from .merge import merge_file
from .profiles import PROFILES
from .rapl import RAPL
from .usage import Usage
from .context import full_context


ENCODERS = ["libx264", "libx265", "libvpx-vp9", "libsvtav1", "vvc"]


def _encode_ref(video: pathlib.Path, profile: str, out: pathlib.Path) -> pathlib.Path:
    """Encode the lossless raw reference video."""
    # colorspace metadata not supported by y4m, nut
    # yuv420p10le not supported by mkv, avi
    # rawvideo failed with flv, obu, thd, vc1, mp4, ismv, dv, mpg, vob, gif, gxf, ogv, webm, y4m
    ref_file = out / f"raw_{video.stem}_{profile}.mkv"
    if ref_file.exists():
        return ref_file
    assert video.is_file(), f"the given file {video} doesn't exists"
    tmp = ref_file.with_stem(f"{ref_file.stem}_in_process")
    with cutcutcodec.read(video) as container:
        stream = container.out_select("video")[0]
        cutcutcodec.write(
            [stream],
            tmp,
            colorspace=cutcutcodec.Colorspace(
                "y'pbpr", PROFILES[profile]["primaries"], PROFILES[profile]["transfer"]
            ),
            # # lossless raw video (no colorspace metadata)
            # streams_settings=[{  # extension .nut for 10 bit support
            #     "encodec": "rawvideo",
            #     "rate": PROFILES[profile]["fps"],
            #     "shape": PROFILES[profile]["shape"][::-1],
            #     "pix_fmt": PROFILES[profile]["pix_fmt"],
            # }],
            # fast lossless compression
            streams_settings=[{  # extension .mkv
                "encodec": "ffv1",
                "rate": PROFILES[profile]["fps"],
                "shape": PROFILES[profile]["shape"][::-1],
                "pix_fmt": PROFILES[profile]["pix_fmt"],
            }],
            # # slow lossless compression
            # streams_settings=[{  # extension .mp4
            #     "encodec": "jpeg2000",
            #     "rate": PROFILES[profile]["fps"],
            #     "shape": PROFILES[profile]["shape"][::-1],
            #     "pix_fmt": PROFILES[profile]["pix_fmt"],
            # }],
            # # very slow
            # streams_settings=[{
            #     "encodec": "libaom-av1",
            #     "rate": PROFILES[profile]["fps"],
            #     "shape": PROFILES[profile]["shape"][::-1],
            #     "pix_fmt": PROFILES[profile]["pix_fmt"],
            #     "options": {"crf": "0", "cpu-used": "8"}, #, "usage": "allintra"},
            # }],
        )
    shutil.move(tmp, ref_file)
    return ref_file


def _get_cmd(
    encoder: str, profile: str, preset: str, crf: float, threads: int
) -> tuple[float, list[str]]:
    """Yield the ffmpeg encode cmd."""
    bit = int(re.search(r"(?P<bit>\d+)le", PROFILES[profile]["pix_fmt"] + "8le")["bit"])

    def libsvtav1_lp(threads: int) -> int:
        """Convert threads in parralel level."""
        # https://gitlab.com/AOMediaCodec/SVT-AV1/-/blob/master/Source/Lib/Globals/enc_handle.c#L749
        return {
            1: 1,
            2: 2, 3: 2, 4: 2, 5: 2,
            6: 3, 7: 3, 8: 3, 9: 3, 10: 3,
            11: 4, 12: 4, 13: 4, 14: 4,
            15: 5, 16: 5,
        }.get(threads, 6)

    match encoder:
        case "libx264":
            return [  # https://ffmpeg.party/x264/
                "libx264",
                "-crf", str(round(crf*51.0/63.0, 1)),
                "-preset", preset,
                "-tune", "ssim",
                "-threads", str(threads),
            ]
        case "libx265":
            return [   # https://x265.readthedocs.io/en/master/cli.html
                "libx265",
                "-crf", str(round(crf*51.0/63.0, 1)),
                "-preset", preset,
                "-tune", "ssim",
                "-x265-params",
                f"frame-threads={threads}:pools={threads}:wpp={1 if threads != 1 else 0}",
            ]
        case "libvpx-vp9":
            return [  # https://wiki.webmproject.org/ffmpeg/vp9-encoding-guide
                "libvpx-vp9",
                "-crf", str(round(crf)),
                "-speed", {"slow": "0", "medium": "1", "fast": "4"}[preset],
                "-tune", "ssim",
                "-row-mt", "1", "-threads", str(threads),
            ]
        case "libsvtav1":
            return [
                "libsvtav1",
                "-crf", str(round(crf)),
                "-preset", {"slow": "4", "medium": "6", "fast": "8"}[preset],
                "-svtav1-params", f"film-grain=0:lp={libsvtav1_lp(threads)}",  # tune=2 broken in v3.0.2
                # "-threads", str(threads),
            ]
        case "vvc":
            return [
                "vvc",
                "-qp", str(round(crf)),
                "-preset", preset,
                "-qpa", "1",
                "-vvenc-params", f"internalbitdepth={bit}",
                "-threads", str(threads),
            ]
        case _:
            raise ValueError(f"the encoder {encoder} is not supported")


def _linspace(start: float, end: float, nbr: int):
    """A linspace omitting the both edges."""
    interval = (end - start) / (nbr + 1)
    return np.linspace(start+interval, start+nbr*interval, nbr).tolist()


@click.command()
@click.argument("video", type=click.Path())
@click.option(
    "-p", "--profile",
    type=click.Choice(list(PROFILES)),
    required=True,
    multiple=True,
    help="The video profile.",
)
@click.option(
    "--preset",
    type=click.Choice(["fast", "medium", "slow"]),
    default=["medium"],
    multiple=True,
    help="The compression effort (default = medium).",
)
@click.option(
    "-e", "--encoder",
    type=click.Choice(list(ENCODERS)),
    default=ENCODERS,
    multiple=True,
    help="The encoder name.",
)
@click.option("-o", "--out", type=click.Path(), help="The output folder.")
@click.option("-r", "--res", type=click.Path(), help="The result json file to be updated.")
@click.option(
    "-n", "--points",
    type=int,
    default=32,
    help="The number of quality point per encoder.",
)
@click.option(
    "-t", "--threads",
    type=int,
    default=[8],
    multiple=True,
    help="The number of threads used by encoders.",
)
@click.option("--psnr/--no-psnr", default=True, help="Compute the PSNR or not.")
@click.option("--ssim/--no-ssim", default=True, help="Compute the SSIM or not.")
@click.option("--uvq/--no-uvq", default=True, help="Compute the UVQ or not.")
@click.option("--vmaf/--no-vmaf", default=True, help="Compute the VMAF or not.")
def main(video: str, out: str = None, res: str = None, **kwargs):
    """Mesure the decoding performance.

    Parameters
    ----------
    video : pathlike
        The video file to be transcoded.
    profile : str
        The video profile including resolution, fps and colorspace.
        Details in ``profiles.py``.
    preset : str, default=medium
        The encoding codec effort, in ``fast``, ``medium``, ``slow``.
    out : str, default=./samples/
        The folder in which to write all transcoded videos.
    res : str, default=./encode.json
        The file containing the results.
    points : int, default=32
        The number of rate per codec.
    threads : int, default=8
        The number of threads
    """
    # verification, preparation
    video = pathlib.Path(video).expanduser()
    out = pathlib.Path(out or "samples").expanduser()
    out.mkdir(parents=True, exist_ok=True)
    res = pathlib.Path(res or pathlib.Path.cwd() / "encode.json")

    # lossless transcode reference file with good shape, rate and colorspace in a raw file
    ref_files = {profile: _encode_ref(video, profile, out) for profile in kwargs["profile"]}

    # iterate on all the parameters
    for profile, threads, preset, crf, encoder in tqdm.tqdm(
        list(itertools.product(
            kwargs["profile"],
            kwargs["threads"],
            kwargs["preset"],
            _linspace(63.0, 0.0, kwargs["points"]),
            kwargs["encoder"],
        )),
        desc="encode",
        dynamic_ncols=True,
        smoothing=1e-2,
        unit="file",
    ):
        encoder_cmd = _get_cmd(encoder, profile, preset, crf, threads)
        file = out / (
            f"{video.stem}_{encoder}_{crf}_{preset}_{threads}_{profile}.mp4"
        )
        if file.exists():
            continue
        tmp = pathlib.Path(tempfile.gettempdir()) / file.name

        # measure IDLE
        idle_duration = 10.0  # duration of the idle measure, in second
        with RAPL(no_fail=True) as idle_rapl, Usage() as idle_usage:
            t_start = time.time()
            time.sleep(idle_duration)
        if idle_rapl is not None:
            del idle_rapl["energy"]
        try:
            idle_wattmeter = g5kpower(platform.node(), t_start, idle_duration)
        except ValueError:
            idle_wattmeter = None
        else:
            del idle_wattmeter["energy"]

        # encode
        with RAPL(no_fail=True) as rapl, Usage() as usage:  # optional measure rapl energy
            t_start = time.time()
            process = subprocess.run([  # start encoding
                "ffmpeg", "-y", "-hide_banner",
                "-i", str(ref_files[profile]), "-c:v", *encoder_cmd,
                str(tmp),
            ], check=True)
            t_stop = time.time()

        # copy result asynchronously
        move_thread = threading.Thread(target=shutil.copy, args=(tmp, file.with_suffix(".tmp")))
        move_thread.start()

        # request to grid5000
        wattmeter_thread = G5kPower(platform.node(), t_start, t_stop-t_start)
        wattmeter_thread.start()

        # run all metrics
        if any(kwargs[m] for m in ["psnr", "ssim", "uvq", "vmaf"]):
            metrics = cutcutcodec.compare(
                ref_files[profile],
                tmp,
                psnr=kwargs["psnr"],
                ssim=kwargs["ssim"],
                uvq=kwargs["uvq"],
                vmaf=kwargs["vmaf"],
            )
        else:
            metrics = {}

        # extract frames informations
        frames = cutcutcodec.core.analysis.ffprobe.get_slices_metadata(tmp)
        frames = [dict(zip(frames[0][0], d)) for d in frames[1].pop().tolist()]

        # merge and write the result
        stats = {
            "args": process.args,
            "cmd": encoder_cmd,
            "compression_ratio": ref_files[profile].stat().st_size / tmp.stat().st_size,
            "crf": crf,
            "encode_context": full_context(),
            "encode_idle_rapl": idle_rapl,
            "encode_idle_usage": idle_usage,
            "encode_idle_wattmeter": idle_wattmeter,
            "encode_rapl": rapl,
            "encode_start": t_start, "encode_stop": t_stop, "encode_duration": t_stop - t_start,
            "encode_threads": threads,
            "encode_usage": usage,
            "encode_wattmeter": wattmeter_thread.get(),
            "encoder": encoder,
            "file": ref_files[profile].stem,
            "frames": frames,
            "preset": preset,
            "profile": profile,
            "size": tmp.stat().st_size,
            "video_duration": float(cutcutcodec.get_duration_video(tmp)),
            **metrics,
        }
        stats["rate"] = 8 * stats["size"] / stats["video_duration"]
        merge_file(res, {file.name: stats})

        # finalize file move
        move_thread.join()
        tmp.unlink()
        shutil.move(file.with_suffix(".tmp"), file)
